// Extract frame from yuv or avi yuv,
// extract frame from second yuv or avi yuv,
// and perform an error image construction
// on it. 

// the error image is to be saved as a
// scuzzy little black and white image
// in the little-known format known as 
// pgm (portable grey map) 

// if you can't read it, install GIMP. It understands them.

//currently only works with yuvs... with headers
// there's a shell script here to put headers onto yuvs though :-)

// Em Tonkin (admin@typodemon.com)
/** 
	Why?

	Because!
 
	May 2002 
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* 

I take this moment to recommend looking at 
Brian W. Kernighan's Programming in C: A Tutorial

I did; it rocks. I mention this only because,
without it, this program would probably be even
worse designed than it is.

*/


struct yuv_format {
  int width; /* Width of YUV image */
  int height; /* Height of YUV image */
  int frames; /* Number of frames in the thing */
  int fps; /* number of frames per second in theory */
  int headerlength; /* How many bytes in the header */
  long int filesize; /* Size of the file */
};

/*
Turns out I don't actually NEED to return a struct... 
//struct yuv_format *get_yuv_format(format,FILE *file) char *format;{ 
//}
*/

int get_y_frame(struct yuv_format *formats,FILE *input_file,int **yframe,int framenumber){

  /* It is fortunate that the 'y' part of YUV is always a fullsize
   frame, except in cases which I'm not dealing with :-)
  Getting the U and the V parts is more problematic. 
  Although in fact, and though I say this far too often in this code,
  there are loads, and LOADS, of different YUV fileformats. You can do 4.2.2
  in lots of different ways; they don't play nice, either.
  I am assuming we convert to MY yuv format using a separate piece of 
  software so that I don't have to support them all right here.
  */


  int yframesize,i,j;

  printf("Getting y frame...\n");
  /* How much data belongs in this frame. */
  yframesize=formats->width*formats->height;
  
  /* I guess we should make /sure/ that filepointer is where we want it... */
  //fseek(input_file,0,formats->headerlength+1);
  rewind(input_file);
  /* If we're actually trying to do a frame other than 0... */
  printf("Fseeking to frame %i\n",framenumber);
  fseek(input_file,(formats->headerlength+1)+(framenumber*formats->width*formats->height*2),0);
  
  /* Fill in the y frame! */
  for(i=0;i<formats->width;i++){
    for(j=0;j<formats->height;j++){
      if(feof(input_file)) return 1;
      yframe[i][j]=getc(input_file);
    }
  }


}

long int get_filesize(FILE *input_file){

  long int i,j,tmpint;
  fseek(input_file,0,0);
  i=ftell(input_file);
  fseek(input_file,0,SEEK_END);
  j=ftell(input_file);
  rewind(input_file);
  tmpint=j-i;
  printf("File length is %i\n",tmpint);
  return tmpint;

}

int get_yuv_format(struct yuv_format *formats,FILE *input_file){
 
  int a,tmpint;

  printf("\n");
  /* Work out the filesize first... */
  formats->filesize=get_filesize(input_file);

  // If this is indeed a YUV file with a little header on it,
  // then this will get the first bit from it -
  // YUV 4:2:2\0	
  // Okay, this should really check if it DOES say YUV and what sort...
  // later
  
  formats->headerlength=23; /* The standard value */

  while((a = getc(input_file)) !='\0') {
    printf("%c",a);
    if(feof(input_file)) break;
  }
  printf("\n");
  
  // If we do this again, you get... width (720), for example
  // On the Bath headers this is a three byte bit followed by another \0 
  
  fscanf(input_file,"%i",&formats->width);
  printf("Width is %i\n",formats->width);
  // NB. The width is followed by a carriage return (\n)
  // This is then followed by the height, eg three byte followed by \0 
  // Warning: Getc considered bloody stupid. However, also 
  // convenient and this is a security unimportant userspace 
  // app... so :-) 
  getc(input_file);	
  tmpint=getc(input_file);
  if(tmpint!='\n') { 
    ungetc(tmpint,input_file);
    formats->headerlength--;
    printf("Missing a return!\n");
  }
  fscanf(input_file,"%i",&formats->height);
  printf("Height is %i\n", formats->height);
  
  // Another carriage return then happens here:\n, followed by 
  // a value frames per second (eg, 30) which is followed by \0
  getc(input_file);
  tmpint=getc(input_file);
  if(tmpint!='\n') {
    printf("Missing a return \n");
    ungetc(tmpint,input_file);
    formats->headerlength--;
  }
  
  fscanf(input_file,"%i",&formats->fps);
  printf("Frames per second : %i\n",formats->fps);
  getc(input_file);
  
  // Yay - that last getc() took in the \0
  // we are now pointing to the start of the movie

  /* Just fill in the number of frames; */
  formats->frames=formats->filesize/(formats->height*formats->width);
  
  
}

void calculate_error(struct yuv_format *formats,int **yframe1,int **yframe2,int visibility, int translation){

  /* Such a short function, and so many args :-) */
  int i,j,tmp;
  printf("\nCalculating difference between yframes\n");
  printf("width is %i and height is %i\n",formats->width,formats->height);
  for(i=0;i<formats->width;i++){
    for(j=0;j<formats->height;j++){
      /* Work out difference between i and j */
      tmp=visibility*(yframe1[i][j]-yframe2[i][j])+translation;
      yframe1[i][j]=tmp;
    }
  }
}

int fwrite_header( FILE *output_file, struct yuv_format *formats)
{
  fprintf( output_file, "%s\n", "P5" );
  fprintf( output_file, "%d %d\n", formats->width, formats->height );
  fprintf( output_file, "%d\n", 255 );
  return( 0 );
}

/* Function to save a array as a pgm file (adapted from Dr. Evans Code) */
void savefile( FILE *output_file, int **yframe, struct yuv_format *formats)  {
  int i,j;
  printf("Formats.height is %i and width is %i\n",formats->height,formats->width);
  fwrite_header(output_file, formats);
  for( i = 0; i < formats->width; i++ ) {
    for( j = 0; j < formats->height; j++ ) {
      fputc(  yframe[i][j], output_file );
    }
  }
  
}


void syntax(){

    printf("\nSyntax is:");
    printf("\n\terrorimage\n\t\tinput filename, input2 filename");		
    printf("\n\t\toutput_name \n");	
    printf("\t Order is important: following these, you may add,\n");
    printf("\t in any order, [--framenumber=n], [--visibility=n],\n");
    printf("\t and [--translation=n]. \n");
    printf("The visibility refers to a multiplication constant \n");
    printf("applied to the error image, the translation refers \n");
    printf("to the 'background shade' of the image (default - 128\n");
    
}


int stringtest(char* needle,char* haystack){
  int returnvalue,i;

  returnvalue=0;

  /* if returnvalue is 0 when it returns, we found it...*/

  /* Checks whether a string starts with another string*/
  /* So this has the behaviour of ~/^needle/ */

  //  printf("Looking for %s in %s",needle,haystack);
  for(i=0;i<strlen(needle);i++){
    if(needle[i]!=haystack[i]){
      returnvalue=1;
      break;
    }
  }
  return(returnvalue);
  
}

int main(int argc, char* argv[])
{

  int **yframe1;
  int **yframe2;
  char a;
  char tempstr[80];
  int end,j,p,k,framenumber;
  struct yuv_format movie1format;
  struct yuv_format movie2format;
  unsigned int length,i;
  FILE *input_file;
  FILE *input_file_2;
  FILE *output_file;
  int tmpint,visibility,translation;

  tmpint=0;
  framenumber=0;
  visibility=2;
  translation=128;


  
  // Because there are some fubared YUVs out there with less \ns
  // than they should have...
  
  
  
  if (argc < 4) {
    syntax();
    return -1;
  }
  
  if(argc>4){
    for(i=4;i<argc;i++){
      if(!stringtest("--visibility=",argv[i])){
	strcpy(tempstr,strchr(argv[i],'=')+1);
	visibility=atoi(tempstr);
	printf("Visibility value set: [%i]\n",visibility);
      } else if(!stringtest("--translation=",argv[i])){
	strcpy(tempstr,strchr(argv[i],'=')+1);
	translation=atoi(tempstr);
	printf("Translation value set: [%i]\n",translation);
      } else if(!stringtest("--framenumber=",argv[i])){
	strcpy(tempstr,strchr(argv[i],'=')+1);
	framenumber=atoi(tempstr);
	printf("Frame value set: [%i]\n",framenumber);
      }
      

    }
  }
  
  
  //all we want is to put a header onto a raw yuv420 sequence.
  // and to turn the stream into 420 if it's 422
  
  
  //open input for reading
  if((input_file=fopen(argv[1],"r"))==NULL){
    
    fprintf(stderr,"Cannot open %s\n",argv[1]);
    return -1;
  }
  
  //open second input file
  if((input_file_2=fopen(argv[2],"r"))==NULL){
    fprintf(stderr,"Cannot open %s\n",argv[2]);
    return -1;
  }

  // open output file
  if((output_file=fopen(argv[3],"wb"))==NULL){
    fprintf(stderr,"Cannot open %s\n",argv[3]);
    return -1;
  }
  
  /* Let us define what sort of movies these are*/

  // XXX if we were smart, it'd check for error codes 
  // and bin out if it returned anything but 0...

  get_yuv_format(&movie1format,input_file);
  get_yuv_format(&movie2format,input_file_2);
  
  if(movie2format.frames>movie1format.frames){
    movie2format.frames=movie1format.frames;
  } else if(movie1format.frames>movie2format.frames){
    movie1format.frames=movie2format.frames;
  }
  
  if(framenumber>movie1format.frames){
    
    printf("Oops, frame number too large. So I set it to the max possible %i\n",movie1format.frames); 
    framenumber=movie1format.frames;
  }

  printf("Check movie formats...\n");

  if(movie1format.width!=movie2format.width || movie2format.height!=movie1format.height){
    printf("Sorry! The movies need to be the same size \n");
    printf("Read documentation for more info on what this software does\n");
    return 1;
  } 

  printf("Movie formats OK\n");
  /* malloc sufficient memory for two sodding great y frames. */
  /* Bear in mind, they are /supposed/ to have the same height/width. */

  printf("Malloc... if it fails here, you got no memory :-) \n");
  yframe1 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );
  yframe2 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );

  for( i = 0; i < movie1format.width; i++ ) {
    yframe1[i] = malloc( movie1format.height*sizeof( int ) );
    yframe2[i] = malloc( movie1format.height*sizeof( int ) );
  }
  printf("Completed malloc... \n");

  get_y_frame(&movie1format,input_file,yframe1,framenumber);
  get_y_frame(&movie2format,input_file_2,yframe2,framenumber);
  
    
  /* This will actually put the difference between the images into yframe1 */

  calculate_error(&movie1format,yframe1,yframe2,visibility,translation);

  printf("Size of height %i,width %i\n",movie1format.height,movie1format.width);
  /* Write yframe1 to disc as an image... */
  

  savefile(output_file,yframe1,&movie1format);

  for(i=0;i<movie1format.width;i++){
    free(yframe1[i]);
    free(yframe2[i]);
  }
  

  fclose(input_file);
  fclose(input_file_2);
  
  return 0;
}

