/** 
	avi to raw conversion software

	Why?

	Because!
 
	May 2002 
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


struct yuv_format {
  int width; /* Width of YUV image */
  int height; /* Height of YUV image */
  int frames; /* Number of frames in the thing */
  int fps; /* number of frames per second in theory */
  int headerlength; /* How many bytes in the header */
  long int filesize; /* Size of the file */
};

static float RGBYUV02990[256], RGBYUV05870[256], RGBYUV01140[256];
static float RGBYUV01684[256], RGBYUV03316[256];
static float RGBYUV04187[256], RGBYUV00813[256];

int get_rgb_frames(struct yuv_format *formats,FILE *input_file,int **rframe, int **gframe, int **bframe, int framenumber){
  int framesize,i,j;
  framesize=formats->width*formats->height;

  /* Fill in the rgb frames! */
  for(i=0;i<formats->width;i++){
    for(j=0;j<formats->height;j++){
      if(feof(input_file)) return 1;
      bframe[i][j]=getc(input_file);
      gframe[i][j]=getc(input_file);
      rframe[i][j]=getc(input_file);
    }
  }


}

int get_y_frame(struct yuv_format *formats,FILE *input_file,int **yframe,int framenumber){

  /* It is fortunate that the 'y' part of YUV is always a fullsize
   frame, except in cases which I'm not dealing with :-)
  Getting the U and the V parts is more problematic. 
  Although in fact, and though I say this far too often in this code,
  there are loads, and LOADS, of different YUV fileformats. You can do 4.2.2
  in lots of different ways; they don't play nice, either.
  I am assuming we convert to MY yuv format using a separate piece of 
  software so that I don't have to support them all right here.
  */


  int yframesize,i,j;

  printf("Getting y frame...\n");
  /* How much data belongs in this frame. */
  yframesize=formats->width*formats->height;
  
   
  /* Fill in the y frame! */
  for(i=0;i<formats->width;i++){
    for(j=0;j<formats->height;j++){
      if(feof(input_file)) return 1;
      yframe[i][j]=getc(input_file);
      getc(input_file);
      getc(input_file);
    }
  }


}


write_yuv_frame(struct yuv_format *formats,FILE *output_file,int **yframe1,int **uframe1,int **vframe1){
  /* Damn, I've forgotten how the YUV thing works... */
  int frames,tmpint,width,height,i,j,k;
  
  // copy y frame
  for(i=0;i<formats->width;i++) {
      for(j=0;j<formats->height;j++){
	putc(yframe1[i][j],output_file);
      }
  }
  
  
  // subsample u
  /* Guessing:

   1 1 1 1 1 1 1 
   1 1 1 1 1 1 1 
   1 1 1 1 1 1 1
   1 1 1 1 1 1 1 

  remove every other row...
  1 1 1 1 1 1 1 

  1 1 1 1 1 1 1

  then remove every other wossname, thingy, column (yawn)
  */

  
  for(j=0; j<height; j+=2)
    {
      for(k=0;k<width;  k+=1)
        {
	  putc(uframe1[k][j],output_file);
	}
    }

  
  // subsample v
  /* Still guessing... :-( */
 
  for(j=0; j<height;j+=2){
    for(k=0;k<width;k+=1){
      putc(vframe1[k][j],output_file);
    }
  }

}

int fwrite_header( FILE *output_file, struct yuv_format *formats)
{
  fprintf( output_file, "%s\n", "P5" );
  fprintf( output_file, "%d %d\n", formats->width, formats->height );
  fprintf( output_file, "%d\n", 255 );
  return( 0 );
}

/* Function to save a array as a pgm file (adapted from Dr. Evans Code) */
void savefile( FILE *output_file, int **yframe, struct yuv_format *formats)  {
  int i,j;
  printf("Formats.height is %i and width is %i\n",formats->height,formats->width);
  fwrite_header(output_file, formats);
  for( i = 0; i < formats->width; i++ ) {
    for( j = 0; j < formats->height; j++ ) {
      fputc(  yframe[i][j], output_file );
    }
  }
  
}

void InitLookupTable()
{
        int i;

        for (i = 0; i < 256; i++) RGBYUV02990[i] = (float)0.2990 * i;
        for (i = 0; i < 256; i++) RGBYUV05870[i] = (float)0.5870 * i;
        for (i = 0; i < 256; i++) RGBYUV01140[i] = (float)0.1140 * i;
        for (i = 0; i < 256; i++) RGBYUV01684[i] = (float)0.1684 * i;
        for (i = 0; i < 256; i++) RGBYUV03316[i] = (float)0.3316 * i;
        for (i = 0; i < 256; i++) RGBYUV04187[i] = (float)0.4187 * i;
        for (i = 0; i < 256; i++) RGBYUV00813[i] = (float)0.0813 * i;
}


void write_yuv_header(FILE *output_file, struct yuv_format *formats){
  /* This much of it appears, and note I say APPEARS, to be working */
  printf("Writing header for YUV 4.2.0 %i by %i image... \n",formats->width,formats->height);
  fprintf( output_file, "YUV 4:2:2");
  putc('\0',output_file);
  fprintf(output_file,"%i",formats->width);
  putc('\0',output_file);
  putc('\n',output_file);
  fprintf(output_file,"%i",formats->height);
  putc('\0',output_file);
  putc('\n',output_file);
  fprintf(output_file,"30");
  putc('\0',output_file);

}



convert_y(struct yuv_format *formats, int **rframe, int **gframe, int **bframe, int **yframe){
  int i,j,tmp;
  
  for( i = 0; i < formats->height; i++ ) {
    for( j = 0; j < formats->width; j++ ) {
      yframe[i][j]=(0.257*rframe[i][j])+(0.504*gframe[i][j])+(0.098*bframe[i][j])+16;
    }
  }
  
}


convert_u(struct yuv_format *formats, int **rframe, int **gframe, int **bframe, int **yframe){
  int i,j,tmp;
  
  for( i = 0; i < formats->height; i++ ) {
    for( j = 0; j < formats->width; j++ ) {
      yframe[i][j]=(0.439*rframe[i][j])+(0.368*gframe[i][j])+(0.071*bframe[i][j])+16;
    }
  }
  
}

convert_v(struct yuv_format *formats, int **rframe, int **gframe, int **bframe, int **yframe){
  int i,j,tmp;
  for( j = 0; j < formats->height; j++ ) {
    for( i = 0; i < formats->width; i++ ) {
      yframe[i][j]=(0.148*rframe[i][j])+(0.291*gframe[i][j])+(0.439*bframe[i][j])+128;


    }
  }
  
}




int stringtest(char* needle,char* haystack){
  int returnvalue,i;

  returnvalue=0;

  /* if returnvalue is 0 when it returns, we found it...*/

  /* Checks whether a string starts with another string*/
  /* So this has the behaviour of ~/^needle/ */
  
  //  printf("Looking for %s in %s",needle,haystack);
  for(i=0;i<strlen(needle);i++){
    if(needle[i]!=haystack[i]){
      returnvalue=1;
      break;
    }
  }
  return(returnvalue);
  
  
  
  
}

int strcpyint(char *mystring,int myint){
  sprintf(mystring,"%i.pgm",myint);
}


int main(int argc, char* argv[])
{
  char Buffer[256];
  char Buffer2[256];
  char a;
  double frames2;
  int frames,end,j,p,k,width,height,fps,headerlen;
  unsigned int length,i;
  FILE *input_file;
  FILE *output_file;
  FILE *output_file2;
  /* need pointer to Buffer for getline */
  int *lineptr;
  int **yframe1;
  int **uframe1;
  int **vframe1;
  int **rframe1;
  int **gframe1;
  int **bframe1;
  struct yuv_format movie1format;
  int tmpint;
  tmpint=0;
  //*lineptr=Buffer[0];


  movie1format.height=486;
  movie1format.width=720;
  //open input for reading
  if((input_file=fopen(argv[1],"r"))==NULL){
    
    fprintf(stderr,"Cannot open %s\n",argv[1]);
    return -1;
  }
  
  //open output
  if((output_file=fopen(argv[2],"wb"))==NULL){
    fprintf(stderr,"Cannot open %s\n",argv[2]);
    return -1;
  }
  
  // This ought to be faster.. apparently not though
  i=ftell(input_file);
  fseek(input_file,0,SEEK_END);
  j=ftell(input_file);
  rewind(input_file);
  tmpint=j-i;
  frames2=movie1format.width*movie1format.height;
  printf("\nvideo stream length=%i, frametest %f\n",tmpint,frames2);
  frames2=tmpint/(3*frames2);
  printf("\n\nWe've found %f frames, width %i, height %i and %i bytes...\n",frames2,movie1format.width,movie1format.height,tmpint);
  printf("\nWe'll round that frames value down...\n");
  /* Yes, I know there are frames going missing. Oh. Dear. Sue me.*/
  movie1format.frames=(int)(frames2);
  printf("after that we have %i frames\n",(int)frames2);

  for(i=0;i<2068;i++){
    j=getc(input_file);
    printf("%i(%c)",j,j);
  }
  printf("Malloc... if it fails here, you got no memory :-) \n");
  yframe1 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );
  uframe1 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );
  vframe1 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );
  rframe1 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );
  gframe1 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );
  bframe1 = ( int ** )malloc( sizeof( int * )*( movie1format.width ) );


  for( i = 0; i < movie1format.width; i++ ) {
    yframe1[i] = malloc( movie1format.height*sizeof( int ) );
    uframe1[i] = malloc( movie1format.height*sizeof( int ) );
    vframe1[i] = malloc( movie1format.height*sizeof( int ) );
    rframe1[i] = malloc( movie1format.height*sizeof( int ) );
    gframe1[i] = malloc( movie1format.height*sizeof( int ) );
    bframe1[i] = malloc( movie1format.height*sizeof( int ) );
          
  }
  printf("\n");
  /* How do I find the end of this goddamn file??? */
  
  fseek(input_file,0,i);
  //get_y_frame(&movie1format,input_file,yframe1,1);
  //get_y_frame(&movie1format,input_file,yframe1,1);
  //get_y_frame(&movie1format,input_file,yframe1,1);
  
  
  //savefile(output_file,yframe1, &movie1format);


  write_yuv_header(output_file,&movie1format);
  InitLookupTable();
  /* Currently assuming that input is raw as that is
     the problem right now */

  /* NB The input is actually BGR... don't ask*/
  for(i=0;i<movie1format.frames;i++){
    /* Get an B */
    /* Get a G */
    /* Get a R  */

    get_rgb_frames(&movie1format,input_file,bframe1, gframe1, bframe1,1);
    
    /* We're going to temporarily do 4:4:4 YUV */
    /* We can cut it down when we save each bit...
       I just have too big a headache as it is */
    
    convert_y(&movie1format,bframe1,gframe1,rframe1,yframe1);
    convert_u(&movie1format,bframe1,gframe1,rframe1,uframe1);
    convert_v(&movie1format,bframe1,gframe1,rframe1,vframe1);
    write_yuv_frame(&movie1format,output_file,yframe1,uframe1,vframe1);

    //open output

     
    /*    printf("Setting filename...\n");
       //strcpy(Buffer,"");
       strcpyint(Buffer,i);
       //strncat(Buffer,".pgm",4);
       printf("Static filename %s\n",Buffer);
       
       if((output_file2=fopen(Buffer,"wb"))==NULL){
       fprintf(stderr,"Cannot open %s\n",argv[2]);
       return -1;
       }  
       savefile(output_file2,yframe1, &movie1format);
       fclose(output_file2);
    */
    
  } /* End of  loop over all frames */


  for( i = 0; i < movie1format.width; i++ ) {
    free(yframe1[i]);
    free(uframe1[i]);
    free(vframe1[i]);
    free(rframe1[i]);
    free(gframe1[i]);
    free(bframe1[i]);
    
  }
  
  free(yframe1);
  free(uframe1);
  free(vframe1);
  free(rframe1);
  free(gframe1);
  free(bframe1);

  fclose(input_file);
  fclose(output_file);
  
  return 0;
}

