#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 

struct header {
  int rows;
  int cols;
};

/******************************* Global Variables ****************************/
int  **image1;
int  **image2;

void printarray( int **temparray, int i, int j, int cols,int rows );
void testtest( int **image );
int testsquare( int **image, int sieve, int **temparray, int i, int j,
		int count, int cols, int rows, int debug );
int fread_header( FILE * fp, struct header * hd );
int fwrite_header( FILE * fp, struct header * hd );
void error( char *msg );
int getline_aux( FILE *file, char *buffer, unsigned int n );
void savefile( FILE * fp, int **image, struct header * hd );
void sobel( int **image2, int cols, int rows );

main( argc, argv)
     int argc;
     char *argv[];
{
  FILE *infile, *outfile1, *outfile2;
  struct header hd1;
  int i, j;
  
  int  sieve_size = 0;
  int  min_sieve_size = 1;
  
  if( argc <= 2 ) {
    printf( "Usage: main: infile [noise-reduction-outfile] edge-detection-outfile\n" );
    exit( 0 );
  }
  
  infile = fopen( argv[1], "rb" );
  if( infile == NULL ) {
    printf( "Can't open input file." );
    exit( 0 );
  }
  
  outfile1 = fopen( argv[2], "wb" );
  if( outfile1 == NULL ) {
    printf( "Can't open output file 1." );
    exit( 0 );
  }
  
  outfile2 = fopen( argv[3], "wb" );
  if( outfile2 == NULL ) {
    printf( "Can't open output file 2." );
    exit( 0 );
  }
  
  fread_header(infile, &hd1);
  image1 = ( int ** )malloc( sizeof( int * )*( hd1.rows ) );
  image2 = ( int ** )malloc( sizeof( int * )*( hd1.rows ) );
  
  for( i = 0; i < hd1.rows; i++ ) {
    image1[i] = malloc( hd1.cols*sizeof( int ) );
    image2[i] = malloc( hd1.cols*sizeof( int ) );
  }

  /* Asking for maximum sieve size from user input*/
  printf( "Enter the maximum sieve size (Enter 0 for no sieving) : " );
  scanf( "%d", &sieve_size ); 
  if( sieve_size != 0 ) {
    printf( "Enter the starting sieve size : " );
    scanf( "%d", &min_sieve_size ); 
    
    printf( "Sieve size going from %d to %d\n", min_sieve_size, sieve_size );
    
  }
  
  for( i = 0; i < hd1.rows; i++ ) { 
    for( j = 0; j < hd1.cols; j++ ) {
      image1[i][j] = getc( infile );
    }
  }

  /* Copy image1 to image2 */
  for( i = 0; i < hd1.rows; i++ ) {
    for( j = 0; j < hd1.cols; j++ ){
      image2[i][j] = image1[i][j];
    }
  }
 
  /* start */

  /* Applying the Sieving noise reduction method */
  if( sieve_size > 0 ) {
    for( i = min_sieve_size;i <= sieve_size; i++ ) {
      reduce_noise( i, image1, image2, hd1.cols, hd1.rows );
    }
    savefile( outfile1, image2, &hd1 );
  }

  printf( "Now applying Sobel Edge Detection... " );
  sobel( image2, hd1.cols, hd1.rows );
  savefile( outfile2, image2, &hd1 );
  printf( "Done.\n" );
  
  /* finish */
  
  
  
  /************************* close all the files ******************************/
  fclose( infile );
  fclose( outfile1 );
  fclose( outfile2 );
  
  return( 0 );
}

/******************************************************************************/

int getline_aux( FILE *file, char *buffer, unsigned int n )
{
  int reading = 0;
  while( !reading ) {
    if( !fgets( buffer, n, file ) )
      return 0;
    reading = ( buffer[0] != '#' );
  }
  return 1;
}


void error( char *msg )
{
  fprintf( stderr, "error in %s\n", msg );
  exit( -1 );
}



int fread_header( FILE * fp, struct header * hd )
{
  int maxgray;
  char buf2[3];
  char buf[256];

  if( !getline_aux( fp, buf, 256 ) || ( sscanf( buf, "%s", buf2 ) != 1 ) ) 
    error( "fread_header: first line" );
  
  if( strcmp( buf2, "P5" ) !=0 )
    error( "fread_header: Not a raw PGM file" );
  
  if( !getline_aux( fp, buf, 256 ) || ( sscanf( buf, "%d %d",
						&( hd->cols ), &( hd->rows ) ) != 2 ) )
    error("fread_header: couldn't read X Y");
  
  if( !getline_aux( fp, buf, 256 ) || ( sscanf( buf, "%d", &maxgray ) != 1 ) )
    error( "fread_header: couldn't read maxgray" );
  
  if( maxgray > 255 )
    error( "fread_header: maxgray > 255" );
  
  return 1;
}

int fwrite_header( FILE * fp, struct header * hd )
{
  fprintf( fp, "%s\n", "P5" );
  fprintf( fp, "%d %d\n", hd->cols, hd->rows );
  fprintf( fp, "%d\n", 255 );
  return( 0 );
}

//------

/******************************************************************************/

/* This is a sieving reduction function which looks around a blob for maxima or
   minima and if the blob is a extrema the extrema colsest to it is assign */

void sieving_reduction(int **image, int **temparray, int **image2,int cols, int rows) {

  int i,j,maximum, minimum,grayshade;
  minimum = 255;
  maximum = 0;

  //		printarray(temparray,0,0,cols,rows);
  //	printf("\n");

  for( i = 1; i < rows - 1; i++ ) {
    for( j = 1; j < cols - 1; j++ ) {
      if(temparray[i][j]==1) {
	/* Search through for the minimum/maximum around each one */
	
	if( image[i - 1][j] > maximum && temparray[i - 1][j] != 1 ) {
	  maximum = image[i - 1][j];
	} 
	
	if( image[i][j + 1] > maximum && temparray[i][j + 1] != 1 ) {
	  maximum = image[i][j + 1];
	}		
	
	if( image[i + 1][j] > maximum && temparray[i + 1][j] != 1 ) {
	  maximum = image[i + 1][j];
	}
	
	if( image[i][j - 1] > maximum && temparray[i][j - 1] != 1 ) {
	  maximum = image[i][j - 1];
	}
	
	/* Dealing with minima */
	
	if( image[i - 1][j] < minimum && temparray[i - 1][j] != 1 ) {
	  minimum = image[i - 1][j];
	}
	
	if( image[i][j + 1] < minimum && temparray[i][j + 1] != 1 ) {
	  minimum = image[i][j + 1];
	}
	
	if( image[i + 1][j] < minimum && temparray[i + 1][j] != 1 ) {
	  minimum = image[i + 1][j];
	}
	
	if( image[i][j - 1] < minimum && temparray[i][j - 1] != 1 ) {
	  minimum = image[i][j - 1];
	}
	
      }
    }
  }
  
  for( i = 1; i < rows - 1; i++ ) {

    for( j = 1; j < cols; j++ ) {

      if( temparray[i][j] == 1 ) {
	if( image[i][j] > maximum ) {	
	  image2[i][j] = maximum;
	  //printf("Replaced %i with max value %i\n",image[i][j],maximum);
	}
	if( image[i][j] < minimum ) {

	  image2[i][j] = minimum;
	  //printf("Replaced %i with min value %i\n",image[i][j],minimum);
	}
	temparray[i][j] = -2;
      }
    }
  }
}

/******************************************************************************/

int reduce_noise(int sieve_size,int **image1, int **image2, int cols,int rows){
  // main function calls the others...
  // Goes through every part of the array looking for conglomerate shapes of the right size, 
  // and then runs a noise reduction algorithm on them...
  
  int i, j, count, tempi, tempj;
  int  **temparray;
  
  for( i = 0; i < rows; i++ ) {
    
    for( j = 0; j < cols; j++ ) {
      
      image1[i][j] = image2[i][j];
    }
  }
  
  temparray = ( int ** )malloc( sizeof( int * ) * ( rows ) );
  
  for( i = 0; i < rows; i++ ) {
    temparray[i] = malloc(cols*sizeof( int ) );
  }
  
  /* Set whole array to zero  except edges which are directly set to -2
     Edges of the images will not be taken into account 
     (if it was important it would not be at the edge */
  for( i = 0; i < rows; i++ ) {
    for( j = 0; j < cols; j++ ) {
      temparray[i][j] = -2;
    }
  }
  for( i = 1; i < rows - 1; i++) {
    for( j = 1; j < cols - 1; j++ ) {
      temparray[i][j] = 0;
    }
  }
  
  /* Cycle left-right, check whether the square is already set to 
     uninteresting in temporary matrix, if not, then check it out with the
     nice recursive function we haven't written yet */
  
  for( i = 1; i < rows - 1; i++ ) {
    for( j = 1; j < cols - 1; j++ ) {
      
      count = testsquare( image1, sieve_size, temparray, i, j, count, cols, rows, 0 );
      if( count != 0 ) {
	/* Aha! A result! Dealing with it */
	
	/* if blob same size as sieve then mark as being correct */
	if( count == sieve_size ) {
	  for( tempi = 1; tempi < rows - 1; tempi++ ) {
	    for( tempj = 1; tempj < cols - 1; tempj++ ) {
	      if( temparray[tempi][tempj] == -1 ) {
		temparray[tempi][tempj] = 1;
	      }
	    }
	  }
	  
	  sieving_reduction(image1, temparray, image2, cols, rows); 
	  
	} else {
	  /* if count not same as sieve, mark as being disregarded */
	  for( tempi = 1; tempi < rows - 1; tempi++ ) {
	    for( tempj = 1; tempj < cols - 1; tempj++ ) {
	      if( temparray[tempi][tempj] == -1 ) {
		temparray[tempi][tempj] = -2;
	      }
	    }
	  }
	}
      } else {
	
	//printf("No result...\n");
      }
      /* If count is zero, we already dealt with this or don't care*/
      
      count = 0;
      //	printf("%i %% of subsweep %i completed\n",j*100/cols,sieve_size);
    }
    printf( "***** %i %% of sweep %i completed\n", i * 100 / rows, sieve_size );
  }
  
  printf( "Completed sieving_reduction for size %i\n", sieve_size );
  
  free(*temparray);
}

/******************************************************************************/

/* This function is called recursively to check if a cell has neighbours of the same value
   it only deal with top, bottom and sides, not diagonals */
int testsquare( int **image, int sieve, int **temparray, int i, int j,
		int count, int cols, int rows, int debug){

  int tempcount = 0;
  int return_value = 0;
  debug++;
  
  if( i > 0 && j > 0 ) {	
    if( i < rows && j < cols ) {
      if( temparray[i][j] == 0 ) {
	/* We never visited this square before. so! Let's handle it properly */
	temparray[i][j] = -1;
	tempcount++;

	if( image[i-1][j] == image[i][j] ) {
	  tempcount += testsquare( image, sieve, temparray, i - 1, j, count, cols, rows, debug );
	}	
	if( image[i][j-1] == image[i][j] ) {
	  tempcount += testsquare( image, sieve, temparray, i, j - 1, count, cols, rows, debug );
	}	
	
	if( image[i+1][j] == image[i][j] ) {
	  tempcount += testsquare( image, sieve, temparray, i + 1, j, count, cols, rows, debug );
	}	
	
	if( image[i][j+1] == image[i][j] ) {
	  tempcount += testsquare( image, sieve, temparray, i, j + 1, count, cols, rows, debug );
	}
	return_value = tempcount;
      } else {
	/* We've been here before, and quite obviously we should be 
	   leaving this square and returning 0! */
	return_value = 0;
      }
    }
  }
  return return_value;
}

/******************************************************************************/

/* This is the sobel edge detector using vertical and horizontal filter */
void sobel( int **image2, int cols, int rows ) {
  
  /* declaring local variables that are needed in the function */
  int i, j, Mx, My, M;
  int temparray[rows][cols];


  /* Main for loop of the sobel edge detector */
  for( i = 1; i < rows- 1; i++ ) {
    for( j = 1; j < cols - 1; j++ ) {
      Mx = - image2[i-1][j-1] - 2 * image2[i][j-1] - image2[i+1][j-1]
	+ image2[i-1][j+1] + 2 * image2[i][j+1] + image2[i+1][j+1];

      My = - image2[i-1][j-1] - 2 * image2[i-1][j] - image2[i-1][j+1]
	+ image2[i+1][j-1] + 2 * image2[i+1][j] + image2[i+1][j+1];

      M = ( abs( Mx ) + abs( My ) ) / 2 ;

      if( M < 30 ) {
	temparray[i][j] = 0;
      } else if( M > 220 ) {
	temparray[i][j] = 255;
      } else temparray[i][j] = M ;
    }
  }


  /* Saving the temporary array to the main one*/
  for ( i = 0; i < rows; i++ ) {
    for ( j = 0; j < cols; j++ ) {
      image2[i][j] = temparray[i][j];
    }
  }
}

/******************************************************************************/

/* Function to save a array as a pgm file (From Dr. Evans Code) */
void savefile( FILE * fp, int **image, struct header * hd)  {
  int i,j;
  fwrite_header(fp, hd);
  for( i = 0; i < hd->rows; i++ ) 
    for( j = 0; j < hd->cols; j++ ) {
      fputc( image[i][j], fp );
    }
  
}

/******************************************************************************/
